<?php
/**
 * zogan - PHP template engine for XML
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Haruki Setoyama <haruki@planewave.org>
 * @copyright Copyright &copy; 2003-2004, Haruki SETOYAMA <haruki@planewave.org>
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @version 0.2- $Id: zogan.elements.php,v 1.8 2004/10/18 09:17:29 haruki Exp $
 * @link http://phpoot.sourceforge.jp/zogan/
 * @package zogan
 * @subpackage _builder
 */
/**
* Base class of template builder for each contents of tag
* This class and its sub classes are accessed only by zogan_parser class.
* @see zogan_parser
* @package zogan
* @subpackage _builder
* @abstract
*/
class zogan_content {
    /**
     * indent string at the beginning of the content
     * @var string
     */
    var $indent = '';

    /**
     * line feed at the end of the content
     * @var string
     */
    var $linefeed = '';

    /**
     * constructor
     */
    function zogan_content($data = '') {}

    /**
     * returns the build PHP script
     * @return string
     */
    function build()
    {
        return '';
    }

    /**
     * returns if this content needs indent and linefeed for pritty print
     * @return string
     */
    function need_indent()
    {
        return false;
    }

    /**
     * returns if this content consites of spaces only or is empty
     * @return string
     */
    function is_spaces_only()
    {
        return false;
    }

    /**
     * returns if this content has variable content for data
     * @return array|false
     */
    function has_variable_content()
    {
        return false;
    }

    /**
     * returns if this class can have alternative content(s)
     * @return array|false
     */
    function is_alternative()
    {
        return false;
    }

    /**
     * double quote for html attribute
     * @access private
     */
    function _double_quote_xml($str)
    {
        return '"' . str_replace(array('"', '<', '>'), array('&quot', '&lt;', '&gt;'), $str) . '"';
    }

    /**
     * single quote for php
     * @access private
     */
    function _single_quote($str)
    {
        return '\'' . str_replace('\'', '\\\'', $str) . '\'';
    }
}

/**
* null object
*/
class zogan_content_null extends zogan_content {

    function zogan_content_null($data = '') {}

    function is_spaces_only()
    {
        return true;
    }
}

/**
* for charactor data
*/
class zogan_content_cdata extends zogan_content {

    /**
     * content string between open and close tag
     * @access private
     */
    var $_content = '';

    function zogan_content_cdata($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent . $this->_content . $this->linefeed;
    }

    function is_spaces_only()
    {
        // tabs or spaces only?
        return (! preg_match('/[^\040\011]/', $this->_content));
    }
}

/**
* default content returns data without modified
*/
class zogan_content_default extends zogan_content {

    /**
     * content string
     * @access private
     */
    var $_content = '';

    function zogan_content_default($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent .
            "<?php echo ".$this->_single_quote($this->_content)." ?>\n"
            . $this->linefeed;
    }
}

/**
* for linefeed
*/
class zogan_content_linefeed extends zogan_content_cdata {
    function zogan_content_linefeed($data)
    {
        parent::zogan_content_cdata($data);
    }

    function is_spaces_only()
    {
        return true;
    }
}

/**
* fake linefeed. This content is for start/end content of root element
*/
class zogan_content_linefeed_fake extends zogan_content_linefeed {
    function zogan_content_linefeed_fake()
    {
        // do nothig
    }

    function build()
    {
        return '';
    }
}

/**
* content for escape ( <! > )
*/
class zogan_content_escape extends zogan_content {

    /**
     * content string
     * @access private
     */
    var $_content;

    function zogan_content_escape($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent . '<!' . $this->_content . '>' . $this->linefeed;
    }
}

/**
* content for pi ( <?[target] ... ?> )
*/
class zogan_content_pi extends zogan_content {

    /**
     * content string
     * @access private
     */
    var $_content;
    /**
     * target string
     * @access private
     */
    var $_target;

    function zogan_content_pi($target, $data)
    {
        $this->_content = $data;
        $this->_target = $target;
    }

    function build()
    {
        $pi = '<?' . $this->_target . $this->_content . '?>';
        $target = $this->_single_quote($this->_target);
        return "<?php if (\$this->_c_pi($target)) { ?>\n"
            .$this->indent . "<?php echo ".$this->_single_quote($pi)." ?>\n" . $this->linefeed
            ."<?php } ?>\n";
    }

    function need_indent()
    {
        return true;
    }
}

/**
* content for removed data
*/
class zogan_content_removed extends zogan_content {

    function need_indent()
    {
        return true;
    }

    function is_spaces_only()
    {
        return true;
    }
}

/**
 * Base class of elements
 * elements are different from other contents in that
 * are possible to have some child contents.
 * @abstract
 */
class zogan_element extends zogan_content {
    /**
     * child contents in this tag
     * @var array of zogan_content
     */
    var $child = '';

    /**
     * information about variable-data tag
     * 'variable-data tag' is the tag with var="{foo}" attribute.
     * @var array
     */
    var $info;

    /**
     * common information in all level of zogan_element
     * 'format':
     * @var array
     */
    var $common;

    /**
     * name of tag
     * @access private
     */
    var $_tag_name;
    /**
     * attribute in the open tag
     * @var array
     * @access private
     */
    var $_attrs;
    /**
     * instructions for template
     * @var array
     * @access private
     * @see zogan_handler::startHandler() in zogan.handler.php
     */
    var $_parsed;
    /**
     * if this tag can have child contents or not
     * @var bool
     * @access private
     */
    var $_empty_tag = false;
    /**
     * if content is to be formated or not
     * @access private
     */
    var $_format = null;
    /**
     * @access private
     */
    var $_linefeed_open_tag = '', $_indent_close_tag = '';
    /**
     * @access private
     */
    var $_content = '';

    /**
     * initialization when the tag is opened
     */
    function open($name, $attrs, $parsed, &$info)
    {
        $this->_tag_name = $name;
        $this->_attrs = $attrs;
        $this->_parsed = $parsed;
        $this->child = array();
    }

    /**
     * parse all the data in $this->child stack
     * and set the compiled PHP script into $this->_content.
     * @access public
     */
    function close($empty_tag)
    {
        $this->_empty_tag = $empty_tag; // bool

        if (count($this->child) == 0
        || $this->_empty_tag == true) {
            $this->_content = '';
        } else {
            $num = count($this->child) -1;

            // for pritty print, make indent
            $this->_indent_close_tag = '';
            if ($this->need_indent()) {
                $indent = $this->_get_indent($num+1);
                if ($indent !== false) {
                    foreach ($indent as $j) {
                        $this->_indent_close_tag .= $this->child[$j]->build();
                        $this->child[$j] = new zogan_content_null;
                    }
                }
            }
            for ($i = $num; $i >= 0; $i--) {
                if ($this->child[$i]->need_indent()) {
                    if (($indent = $this->_get_indent($i)) !== false
                    && ($lf = $this->_get_linefeed($i)) !== false) {
                        foreach ($indent as $j) {
                            $this->child[$i]->indent .= $this->child[$j]->build();
                            $this->child[$j] = new zogan_content_null;
                        }
                        foreach ($lf as $j) {
                            $this->child[$i]->linefeed .= $this->child[$j]->build();
                            $this->child[$j] = new zogan_content_null;
                        }
                    }
                }
            }
            $this->_linefeed_open_tag = '';
            $lf = $this->_get_linefeed(-1);
            if ($lf !== false) {
                foreach ($lf as $j) {
                    $this->_linefeed_open_tag .= $this->child[$j]->build();
                    $this->child[$j] = new zogan_content_null;
                }
            }

            // parsing in consideration of altarnative feature
            $this->_content = '';
            for ($i = 0; $i <= $num; $i++) {
                if ($this->child[$i]->has_variable_content() !== true) {
                    $this->_content .= $this->child[$i]->build();
                } else {
                    $level      = $this->child[$i]->info['level'];
                    $_each      = '$_each' . $level;
                    $_valprev   = '$_val'  . ($level -1);
                    $_valcurr   = '$_val'  . $level;
                    $_num       = '$_num'  . $level;

                    $target = $this->_get_c_val_func($_valprev, $this->child[$i]->info['name']);

                    $this->_content
                        .= '<?php '
                        . "$_each = \$this->_c_each($target); "
                        . "$_num = 0; "
                        . "foreach ($_each as $_valcurr) { ";

                    if ($this->child[$i]->is_alternative() !== false) {
                        $i = $this->_parse_alternative($i);
                    } else {
                        $this->_content .= $this->child[$i]->build();
                    }

                    $this->_content .= "} ?>\n";
                }
            }
        }
    }

    /**
     * get white spaces for indent
     * @access private
     * @return array index of contents for indent
     */
    function _get_indent($i)
    {
        $indent = array();
        for ($j = $i - 1; $j >= 0; $j--){
            if (is_a($this->child[$j], 'zogan_content_linefeed')) {
                break;
            }
            elseif ($this->child[$j]->is_spaces_only()) {
                $indent[] = $j;
                continue;
            }
            else {
                return false;
            }
        }
        if ($j < 0) {
            return false;
        }
        return $indent;
    }

    /**
     * get white spaces and linefeed after tag for pretty print
     * @access private
     * @return false|array index of contents
     */
    function _get_linefeed($i)
    {
        $num = count($this->child) -1;
        $lf = array();
        for ($j = $i + 1; $j <= $num; $j++){
            if (is_a($this->child[$j], 'zogan_content_linefeed')) {
                $lf[] = $j;
                break;
            }
            elseif ($this->child[$j]->is_spaces_only()) {
                $lf[] = $j;
                continue;
            }
            else {
                return false;
            }
        }
        if ($j > $num) {
            return false;
        }
        return $lf;
    }

    /**
     * parse the tag with alternative
     * @access private
     * @return string
     */
    function _parse_alternative($i)
    {
        $num = count($this->child);
        $level = $this->child[$i]->info['level'];
        $alts = array($i);
        $sign = $this->child[$i]->is_alternative();
        for ($i++; $i < $num; $i++){
            if ($this->child[$i]->is_alternative() === $sign) {
                $alts[] = $i;
                continue;
            }
            elseif ($this->child[$i]->is_spaces_only() == true) {
                continue;
            }
            break;
        }

        $_num  = '$_num'  . $level;
        $_alts = count($alts);

        $_j = 0;
        foreach ($alts as $h) {
            $this->_content
                .= "if ($_num%$_alts == $_j) { "
                . $this->child[$h]->build()
                . "continue; } ";
            $_j++;
        }

        return end($alts);
    }

    /**
     * returns PHP code with phpoot::_c_val* method
     * @return stinrg
     * @access private
     */
    function _get_c_val_func($val, $var_name_array)
    {
        if (count($var_name_array) > 1) {
            $_var = 'array(\''.implode('\',\'', $var_name_array).'\')';
            return "\$this->_c_val_array($val, $_var)";
        } else {
            $_var = '\''.$var_name_array[0].'\'';
            return "\$this->_c_val($val, $_var)";
        }
    }

    /**
     * returns build tag
     * @return array array($open_tag, $close_tag)
     * @access private
     */
    function _build_tag()
    {
        // omit open & close tag?
        if (@$this->_parsed[ZOGAN_OMIT_TAG] === true) {
            $open_tag = '';
            $close_tag = '';
            return;
        }

        // open tag
        $open_tag = '<' . $this->_tag_name;
        $_val = '$_val' . $this->_build_tag_get_level();

        foreach ($this->_attrs as $key => $value) {

            if (is_array($value)) {
            // variable attr
                $_key = $value['key'];
                $_default = $this->_single_quote($value['default']);
                $get_val = $this->_get_c_val_func($_val, $value['var']);
                if ($value['func'] != '') {
                    $get_val = sprintf("\$this->_c_func(%s, '%s', '%s')",
                                    $get_val, $value['func'], $value['param']);
                }
                $open_tag .= "<?php \$this->_c_attr('$_key', $get_val, $_default); ?>\n";
            }
            elseif (is_string($value)) {
            // static attr
                $open_tag .= ' ' . $key . '=' . $this->_double_quote_xml($value);
            }
            elseif ($value === true) {
            // var:cont attr
                $_val = '$_val' . $this->info['level'];
            }
        }

        if ($this->_empty_tag == true) {
            $open_tag  .= ' />';
            $close_tag  = '';
        } else {
            $open_tag  .= '>';
            $close_tag  = '</' . $this->_tag_name . '>';
        }

        return array($open_tag, $close_tag);
    }

    function _build_tag_get_level()
    {
        return $this->info['level'];
    }

    /**
     * returns namespace information for ZOGAN
     * @return array
     */
    function getNamespace()
    {
        return $this->_parsed[ZOGAN_NAMESPACE];
    }

    /**
     * returns name of tag
     * @return array
     */
    function getTagname()
    {
        return $this->_tag_name;
    }

    /**
     * build when ZOGAN_OMIT=='first'
     * @return string
     */
    function _omit_first($built)
    {
        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $_num   = '$_num'  . $this->info['level'];
            return "<?php if ($_num>1) { ?>\n". $built . "<?php } ?>\n";
        } else {
            return $built;
        }
    }
}

/**
 * Parser for root level of the stack
 *
 * @package zogan
 */
class zogan_element_root extends zogan_element {

    function zogan_element_root()
    {
        $this->child = array();
        $this->child[] = new zogan_content_linefeed_fake;
        $this->_parsed[ZOGAN_NAMESPACE]
            = array(ZOGAN_PREFIX => ZOGAN_NAMESPACE);
        $this->info
            = array('level' => 0, 'nests' => false);
    }

    function close($close_tag)
    {
        // do nothing. In fact, this method is not called.
        return;
    }

    function build()
    {
        $this->child[] = new zogan_content_linefeed_fake;
        parent::close(false);
        return $this->_content;
    }
}


/**
 * Parser for tags without variable attribute
 *
 * @package zogan
 */
class zogan_element_static extends zogan_element {

    function open($name, $attrs, $parsed, &$info)
    {
        $this->info =& $info;
        parent::open($name, $attrs, $parsed, $info);
    }

    function build()
    {

        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;
        $built = $open_tag . $this->_linefeed_open_tag
                . $this->_content
                . $this->_indent_close_tag . $close_tag;
        $built = $this->_omit_first($built);
        return $built;
    }

    function need_indent()
    {
        return isset($this->_parsed[ZOGAN_OMIT]);
    }
}

/**
 * Parser for variable tag
 * variable tag means the one with var attribute.
 *
 * @package zogan
 */
class zogan_element_variable extends zogan_element {

    function open($name, $attrs, $parsed, &$info)
    {
        $this->info['level'] = (int)$info['level'] +1;
        $this->info['nests'] = false;
        $info['nests'] = true;
        $this->info['name']  = $parsed[ZOGAN_CONTENT]['var'];
        $this->info['func']  = $parsed[ZOGAN_CONTENT]['func'];
        $this->info['param'] = $parsed[ZOGAN_CONTENT]['param'];
        parent::open($name, $attrs, $parsed, $info);
    }

    function build()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_var       = end($this->info['name']);
        $_valcurr   = '$_val'  . $this->info['level'];
        $_num       = '$_num'  . $this->info['level'];
        $escape     = 'true';
        $built      = '';

        if ($this->info['func'] != '') {
            $built .= sprintf("$_valcurr = \$this->_c_func($_valcurr, '%s', '%s'); ",
                            $this->info['func'], $this->info['param']);
            $escape = 'false';
        }

        $built
            .= "if ($_valcurr === null || $_valcurr === false) { "
            . "if (!\$this->_quiet) { ?>\n"
            . $this->indent . "<!-- no data in '$_var' -->" . $this->linefeed
            . '<?php } } else { '
            . "$_num++; ";

        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $built .= "if ($_num>1) { ";
        }

        if (! $this->info['nests']) {
        // not nested
            $built
               .= "?>\n"
                . $open_tag
                . '<?php '
                . "if ($_valcurr === true) { ?>\n"
                . $this->_linefeed_open_tag . $this->_content . $this->_indent_close_tag
                . "<?php } "
                . "else { \$this->_c_content($_valcurr, '$_var', $escape); } "
                . "?>\n"
                . $close_tag;
        } else {
        // nested
            $built
                .= "\$this->_c_hash($_valcurr, '$_var'); ?>\n"
                 . $open_tag . $this->_linefeed_open_tag
                 . $this->_content
                 . $this->_indent_close_tag . $close_tag;
        }

        $built
            .= '<?php ';

        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $built .= '} ';
        }

        $built
            .= '} ';

        return $built;
    }

    function need_indent()
    {
        return true;
    }

    function is_alternative()
    {
        if (@$this->_parsed[ZOGAN_ALTERNATON] === true) {
            return array($this->_tag_name, $this->info['name']);
        } else {
            return false;
        }
    }

    function has_variable_content()
    {
        return true;
    }

    function _build_tag_get_level()
    {
        return ($this->info['level'] > 0) ? $this->info['level']-1 : 0;
    }
}

/**
 * Parser for tags with '#num' of attribute for variable
 * which shows the number of iteration.
 * @package zogan
 */
class zogan_element_num extends zogan_element_static {

    function build()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_num = '$_num' . $this->info['level'];
        $built = $open_tag . $this->_linefeed_open_tag
                . "<?php echo $_num; ?>\n"
                . $this->_indent_close_tag . $close_tag;
        $built = $this->_omit_first($built);
        return $built;
    }
}

/**
 * include other template
 * @package zogan
 * @todo when non existant file?
 */
class zogan_element_inc extends zogan_element_static {

    function build()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag  = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_valcurr = '$_val' . $this->info['level'];
        $_zogan   = '$_ppt' . $this->info['level'];
        $_incfile = $this->_parsed[ZOGAN_CONTENT]['arg'];

        $built = $open_tag . $this->_linefeed_open_tag
           . "<?php "
           . "$_zogan = zogan_clone_obj(\$this); "
           . $_zogan."->_c_change_basename(".$this->_single_quote($_incfile)."); "
           . $_zogan."->display($_valcurr); "
           . "unset($_zogan); ?>\n"
           . $this->_indent_close_tag. $close_tag;

        $built = $this->_omit_first($built);
        return $built;
    }
}

/**
 * omitted element
 */
class zogan_element_omit extends zogan_element_static {
    function build()
    {
        return '';
    }

    function need_indent()
    {
        return true;
    }

    function is_spaces_only()
    {
        return true;
    }
}

?>